# Copyright 2018 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from inspect import signature
from enum import IntEnum
from abc import abstractmethod
from ovos_utils.log import LOG
from ovos_utils import get_mycroft_root
from ovos_utils.waiting_for_mycroft.base_skill import MycroftSkill

try:
    from mycroft.skills.common_play_skill import CommonPlaySkill as _CommonPlaySkill
except ImportError:
    import sys

    MYCROFT_ROOT_PATH = get_mycroft_root()
    if MYCROFT_ROOT_PATH is not None:
        sys.path.append(MYCROFT_ROOT_PATH)
        from mycroft.skills.common_play_skill import CommonPlaySkill as _CommonPlaySkill
    else:
        LOG.error("Could not find mycroft root path")
        raise ImportError

# implementation of
# https://github.com/MycroftAI/mycroft-core/pull/2660


class CPSMatchLevel(IntEnum):
    EXACT = 1
    MULTI_KEY = 2
    TITLE = 3
    ARTIST = 4
    CATEGORY = 5
    GENERIC = 6


class CPSTrackStatus(IntEnum):
    DISAMBIGUATION = 1  # not queued for playback, show in gui
    PLAYING = 20  # Skill is handling playback internally
    PLAYING_AUDIOSERVICE = 21  # Skill forwarded playback to audio service
    PLAYING_GUI = 22  # Skill forwarded playback to gui
    PLAYING_ENCLOSURE = 23  # Skill forwarded playback to enclosure
    QUEUED = 30  # Waiting playback to be handled inside skill
    QUEUED_AUDIOSERVICE = 31  # Waiting playback in audio service
    QUEUED_GUI = 32  # Waiting playback in gui
    QUEUED_ENCLOSURE = 33  # Waiting for playback in enclosure
    PAUSED = 40  # media paused but ready to resume
    STALLED = 60  # playback has stalled, reason may be unknown
    BUFFERING = 61  # media is buffering from an external source
    END_OF_MEDIA = 90  # playback finished, is the default state when CPS loads


class CPSMatchType(IntEnum):
    GENERIC = 1
    MUSIC = 2
    VIDEO = 3
    AUDIOBOOK = 4
    GAME = 5
    PODCAST = 6
    RADIO = 7
    NEWS = 8
    TV = 9
    MOVIE = 10
    TRAILER = 11
    ADULT = 12
    VISUAL_STORY = 13


class CommonPlaySkill(MycroftSkill, _CommonPlaySkill):
    """ To integrate with the common play infrastructure of Mycroft
    skills should use this base class and override the two methods
    `CPS_match_query_phrase` (for checking if the skill can play the
    utterance) and `CPS_start` for launching the media.

    The class makes the skill available to queries from the
    mycroft-playback-control skill and no special vocab for starting playback
    is needed.
    """
    def __init__(self, name=None, bus=None):
        super().__init__(name, bus)

        self.supported_media = [CPSMatchType.GENERIC, CPSMatchType.MUSIC]
        # NOTE: derived skills will likely want to override this list,
        # for backwards compatibility MUSIC is supported as a default

    def __handle_play_query(self, message):
        """Query skill if it can start playback from given phrase."""
        search_phrase = message.data["phrase"]
        media_type = message.data.get("media_type", CPSMatchType.GENERIC)

        if media_type not in self.supported_media:
            return

        # First, notify the requestor that we are attempting to handle
        # (this extends a timeout while this skill looks for a match)
        self.bus.emit(message.response({"phrase": search_phrase,
                                        "skill_id": self.skill_id,
                                        "searching": True}))

        # Now invoke the CPS handler to let the skill perform its search
        if len(signature(self.CPS_match_query_phrase).parameters) == 2:
            result = self.CPS_match_query_phrase(search_phrase, media_type)
        else:
            # needed for old skills which are not expecting media_type
            # TODO remove in next major release
            result = self.CPS_match_query_phrase(search_phrase)

        if result:
            match = result[0]
            level = result[1]
            callback = result[2] if len(result) > 2 else None
            confidence = self.__calc_confidence(match, search_phrase, level)
            self.bus.emit(message.response({"phrase": search_phrase,
                                            "skill_id": self.skill_id,
                                            "callback_data": callback,
                                            "service_name": self.spoken_name,
                                            "conf": confidence}))
        else:
            # Signal we are done (can't handle it)
            self.bus.emit(message.response({"phrase": search_phrase,
                                            "skill_id": self.skill_id,
                                            "searching": False}))

    def __handle_play_start(self, message):
        """Bus handler for starting playback using the skill."""
        if message.data["skill_id"] != self.skill_id:
            # Not for this skill!
            return
        phrase = message.data["phrase"]
        data = message.data.get("callback_data")

        # Stop any currently playing audio
        if self.audioservice.is_playing and \
                message.data.get("stop_audio", True):
            # "stop_audio" is used by playback control, it avoids calling stop
            # if query was interpreted as a "resume" command
            self.audioservice.stop()

            self.bus.emit(message.forward("mycroft.stop"))

        # Save for CPS_play() later, e.g. if phrase includes modifiers like
        # "... on the chromecast"
        self.play_service_string = phrase

        # Invoke derived class to provide playback data
        self.CPS_start(phrase, data)

    ######################################################################
    # Abstract methods
    # All of the following must be implemented by a skill that wants to
    # act as a CommonPlay Skill
    @abstractmethod
    def CPS_match_query_phrase(self, phrase, media_type):
        """Analyze phrase to see if it is a play-able phrase with this skill.

        Arguments:
            phrase (str): User phrase uttered after "Play", e.g. "some music"

        Returns:
            (match, CPSMatchLevel[, callback_data]) or None: Tuple containing
                 a string with the appropriate matching phrase, the PlayMatch
                 type, and optionally data to return in the callback if the
                 match is selected.
        """
        # Derived classes must implement this, e.g.
        #
        # if phrase in ["Zoosh"]:
        #     return ("Zoosh", CPSMatchLevel.Generic, {"hint": "music"})
        # or:
        # zoosh_song = find_zoosh(phrase)
        # if zoosh_song and "Zoosh" in phrase:
        #     # "play Happy Birthday in Zoosh"
        #     return ("Zoosh", CPSMatchLevel.MULTI_KEY, {"song": zoosh_song})
        # elif zoosh_song:
        #     # "play Happy Birthday"
        #     return ("Zoosh", CPSMatchLevel.TITLE, {"song": zoosh_song})
        # elif "Zoosh" in phrase
        #     # "play Zoosh"
        #     return ("Zoosh", CPSMatchLevel.GENERIC, {"cmd": "random"})
        return None



import random
import json


class MarkovChain:
    """
    Simple Markov Chain Class
    """
    START_OF_SEQ = "~"
    END_OF_SEQ = "[END]"

    def __init__(self, order=1, pad=True, records=None):
        """
        Initialise Markov chain
        :param order: int - number of tokens to consider a state
        :param pad: bool - whether to pad training strings with start/end tokens
        """
        self.order = order
        self.pad = pad
        self.records = {} if records is None else records

    def add_tokens(self, tokens):
        """
        Adds a list of tokens to the markov chain
        :param tokens: list of tokens
        :return: None
        """
        if self.pad:
            tokens = [self.START_OF_SEQ] * self.order + tokens + [
                self.END_OF_SEQ]

        for i in range(len(tokens) - self.order):
            current_state = tuple(tokens[i:i + self.order])
            next_state = tokens[i + self.order]
            self.add_state(current_state, next_state)

    def add_state(self, current_state, next_state):
        """
        Updates the weight of the transition from current_state to next_state
        with a single observation.
        :param current_state: tuple - current state
        :param next_state: token - the next observed token
        :return: None
        """
        if current_state not in self.records.keys():
            self.records[current_state] = dict()

        if next_state not in self.records[current_state].keys():
            self.records[current_state][next_state] = 0

        self.records[current_state][next_state] += 1

    def generate_sequence(self, n=100, initial_state=None):
        """
        Generates a sequence of tokens from the markov chain, starting from
        initial_state. If initial state is empty, and pad is false it chooses an
        initial state at random. If pad is true,
        :param n: int - The number of tokens to generate
        :param initial_state: starting state of the generator
        :return: list of generated tokens
        """

        if initial_state is None:
            if self.pad:
                sequence = [START_OF_SEQ] * self.order
            else:
                sequence = list(random.choice(self.records.keys()))
        else:
            sequence = initial_state[:]

        for i in range(n):
            current_state = tuple(sequence[-self.order:])
            next_token = self.sample(current_state)
            sequence.append(next_token)

            if next_token == self.END_OF_SEQ:
                return sequence

        return sequence

    def sample(self, current_state, random_on_fail=True):
        """
        Generates a random next token, given current_state
        :param current_state: tuple - current_state
        :return: token
        """
        if current_state not in self.records.keys() and random_on_fail:
            current_state = random.choice(self.records.keys())

        possible_next = self.records[current_state]
        n = sum(possible_next.values())

        m = random.randint(0, n)
        count = 0
        for k, v in possible_next.items():
            count += v
            if m <= count:
                return k

    def save(self, filename):
        """
        Saves Markov chain to filename
        :param filename: string - where to save chain
        :return: None
        """
        with open(filename, "w") as f:
            m = {
                "order": self.order,
                "pad": self.pad,
                "records": {str(k): v for k, v in self.records.items()}
            }
            json.dump(m, f)

    @staticmethod
    def load(filename):
        """
        Loads Markov chain from json file
        DUE TO USE OF EVAL
        DO NOT RUN THIS ON UNTRUSTED FILES
        :param filename:
        :return: MarkovChain
        """
        with open(filename, "r") as f:
            raw = json.load(f)

        mc = MarkovChain(
            raw["order"],
            raw["pad"],
            {eval(k): v for k, v in raw["records"].items()}
        )

        return mc



from phoneme_guesser import guess_phonemes as _guess_phonemes, \
    get_phonemes as _get_phonemes

# Backwards compat, TODO deprecate?


def guess_phonemes(word, lang="en-us"):
    return _guess_phonemes(word, lang)


def get_phonemes(name, lang="en-us"):
    return _get_phonemes(name, lang)


if __name__ == "__main__":

    words = ["hey mycroft", "hey chatterbox", "alexa", "siri", "cortana"]
    for w in words:
        print(w, get_phonemes(w))

    """
    hey mycroft HH EY1 . M Y K R OW F T
    hey chatterbox HH EY1 . CH AE T EH R B OW K S
    alexa AH0 L EH1 K S AH0
    siri S IH1 R IY0
    cortana K OW R T AE N AE
    """


import re
from os.path import join, exists
from itertools import chain
from ovos_utils.log import LOG
from ovos_utils import get_mycroft_root
from ovos_utils.waiting_for_mycroft.skill_gui import SkillGUI
try:
    from mycroft.skills.mycroft_skill import MycroftSkill as _MycroftSkill
    from mycroft.skills.fallback_skill import FallbackSkill as _FallbackSkill
    from mycroft.skills.skill_data import read_vocab_file
    from mycroft.util import resolve_resource_file
except ImportError:
    import sys
    MYCROFT_ROOT_PATH = get_mycroft_root()
    if MYCROFT_ROOT_PATH is not None:
        sys.path.append(MYCROFT_ROOT_PATH)
        from mycroft.skills.mycroft_skill import MycroftSkill as _MycroftSkill
        from mycroft.skills.fallback_skill import FallbackSkill as _FallbackSkill
        from mycroft.skills.skill_data import read_vocab_file
        from mycroft.util import resolve_resource_file
    else:
        LOG.error("Could not find mycroft root path")
        raise ImportError


class MycroftSkill(_MycroftSkill):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # https://github.com/MycroftAI/mycroft-core/pull/2683
        self.gui = SkillGUI(self)

    # https://github.com/MycroftAI/mycroft-core/pull/2675
    def voc_match(self, utt, voc_filename, lang=None, exact=False):
        """Determine if the given utterance contains the vocabulary provided.
        Checks for vocabulary match in the utterance instead of the other
        way around to allow the user to say things like "yes, please" and
        still match against "Yes.voc" containing only "yes". The method first
        checks in the current skill's .voc files and secondly the "res/text"
        folder of mycroft-core. The result is cached to avoid hitting the
        disk each time the method is called.
        Arguments:
            utt (str): Utterance to be tested
            voc_filename (str): Name of vocabulary file (e.g. 'yes' for
                                'res/text/en-us/yes.voc')
            lang (str): Language code, defaults to self.long
            exact (bool): comparison using "==" instead of "in"
        Returns:
            bool: True if the utterance has the given vocabulary it
        """
        lang = lang or self.lang
        cache_key = lang + voc_filename
        if cache_key not in self.voc_match_cache:
            # Check for both skill resources and mycroft-core resources
            voc = self.find_resource(voc_filename + '.voc', 'vocab')
            if not voc:  # Check for vocab in mycroft core resources
                voc = resolve_resource_file(join('text', lang,
                                                 voc_filename + '.voc'))

            if not voc or not exists(voc):
                raise FileNotFoundError(
                    'Could not find {}.voc file'.format(voc_filename))
            # load vocab and flatten into a simple list
            vocab = read_vocab_file(voc)
            self.voc_match_cache[cache_key] = list(chain(*vocab))
        if utt:
            if exact:
                # Check for exact match
                return any(i.strip() == utt
                           for i in self.voc_match_cache[cache_key])
            else:
                # Check for matches against complete words
                return any([re.match(r'.*\b' + i + r'\b.*', utt)
                            for i in self.voc_match_cache[cache_key]])
        else:
            return False


class FallbackSkill(MycroftSkill, _FallbackSkill):
    """ """



from ovos_utils.sound import play_mp3
from ovos_utils.lang.detect import detect_lang
from ovos_utils.lang import get_tts
from ovos_utils.log import LOG

import requests

try:
    from google_trans_new import google_translator
except ImportError:
    google_translator = None

import logging

logging.getLogger("hpack").setLevel("ERROR")
logging.getLogger("chardet").setLevel("ERROR")


def translate_text(text, lang="en-us", source_lang=None):
    tx = translate_apertium(text, lang, source_lang)
    if not tx:
        LOG.warning("Falling back to google translate")
        return translate_google(text, lang, source_lang)
    return tx


def translate_to_mp3(sentence, lang="en-us", mp3_file=None):
    if detect_lang(sentence) != lang.split("-")[0]:
        sentence = translate_text(sentence, lang)
    return get_tts(sentence, lang, mp3_file)


def say_in_language(sentence, lang="en-us", mp3_file=None):
    mp3_file = translate_to_mp3(sentence, lang, mp3_file)
    play_mp3(mp3_file)
    return sentence


def translate_apertium(text, lang="en-us", source_lang=None):
    lang_pair = lang
    if source_lang:
        lang_pair = source_lang + "|" + lang
    r = requests.get("https://www.apertium.org/apy/translate",
                     params={"q": text, "langpair": lang_pair}).json()
    if r.get("status", "") == "error":
        LOG.error(r["explanation"])
        return None
    return r["responseData"]["translatedText"]


def translate_google(text, lang="en-us", source_lang=None):
    if google_translator is None:
        LOG.debug("run pip install google_trans_new")
        raise ImportError("google_trans_new not installed")
    translator = google_translator()
    lang = lang.split("-")[0]
    if source_lang:
        source_lang = source_lang.split("-")[0]
        tx = translator.translate(text, lang_src=source_lang, lang_tgt=lang)
    else:
        tx = translator.translate(text,  lang_tgt=lang)
    return tx




from ovos_utils.log import LOG
from ovos_utils.enclosure import enclosure2rootdir, detect_enclosure
from ovos_utils.enclosure import MycroftEnclosures
from ovos_utils.json_helper import merge_dict, load_commented_json
from os.path import isfile, exists, expanduser, join, dirname, isdir
from os import makedirs
import json

MYCROFT_SYSTEM_CONFIG = "/etc/mycroft/mycroft.conf"
MYCROFT_USER_CONFIG = join(expanduser("~"), ".mycroft", "mycroft.conf")


def get_config_fingerprint():
    conf = read_mycroft_config()
    listener_conf = conf.get("listener", {})
    skills_conf = conf.get("skills", {})
    return {
        "enclosure": conf.get("enclosure", {}).get("platform"),
        "data_dir": conf.get("data_dir"),
        "msm_skills_dir": skills_conf.get("msm", {}).get("directory"),
        "ipc_path": conf.get("ipc_path"),
        "input_device_name": listener_conf.get("device_name"),
        "input_device_index": listener_conf.get("device_index"),
        "default_audio_backend": conf.get("Audio", {}).get("default-backend"),
        "priority_skills": skills_conf.get("priority_skills"),
        "backend_url": conf.get("server", {}).get("url")
    }


def read_mycroft_config():
    conf = LocalConf(None)
    conf.merge(MycroftDefaultConfig())
    conf.merge(MycroftSystemConfig())
    conf.merge(MycroftUserConfig())
    return conf


def update_mycroft_config(config, path=None):
    if path is None:
        conf = MycroftUserConfig()
    else:
        conf = LocalConf(path)
    conf.merge(config)
    conf.store()
    return conf


# TODO use json_database.JsonStorage
class LocalConf(dict):
    """
        Config dict from file.
    """
    allow_overwrite = True

    def __init__(self, path):
        super(LocalConf, self).__init__()
        self.path = path
        if self.path:
            self.load_local(self.path)

    def load_local(self, path):
        """
            Load local json file into self.

            Args:
                path (str): file to load
        """
        path = expanduser(path)
        if exists(path) and isfile(path):
            try:
                config = load_commented_json(path)
                for key in config:
                    self.__setitem__(key, config[key])

                LOG.debug("Configuration {} loaded".format(path))
            except Exception as e:
                LOG.error("Error loading configuration '{}'".format(path))
                LOG.error(repr(e))
        else:
            LOG.debug("Configuration '{}' not defined, skipping".format(path))

    def reload(self):
        self.load_local(self.path)

    def store(self, path=None):
        """
            store the configuration locally.
        """
        path = path or self.path
        if not path:
            LOG.warning("config path not set, updating user config!!")
            update_mycroft_config(self)
            return
        path = expanduser(path)
        if not isdir(dirname(path)):
            makedirs(dirname(path))
        with open(path, 'w', encoding="utf-8") as f:
            json.dump(self, f, indent=4, ensure_ascii=False)

    def merge(self, conf):
        merge_dict(self, conf)
        return self


class ReadOnlyConfig(LocalConf):
    """ read only  """
    def __init__(self, path, allow_overwrite=False):
        super().__init__(path)
        self.allow_overwrite = allow_overwrite

    def reload(self):
        old = self.allow_overwrite
        self.allow_overwrite = True
        super().reload()
        self.allow_overwrite = old

    def __setitem__(self, key, value):
        if not self.allow_overwrite:
            raise PermissionError
        super().__setitem__(key, value)

    def __setattr__(self, key, value):
        if not self.allow_overwrite:
            raise PermissionError
        super().__setattr__(key, value)

    def merge(self, conf):
        if not self.allow_overwrite:
            raise PermissionError
        super().merge(conf)

    def store(self, path=None):
        if not self.allow_overwrite:
            raise PermissionError
        super().store(path)


class MycroftUserConfig(LocalConf):
    def __init__(self):
        path = MYCROFT_USER_CONFIG
        enclosure = detect_enclosure()
        if enclosure == MycroftEnclosures.MARK1 or \
                enclosure == MycroftEnclosures.OLD_MARK1:
            path = "/home/mycroft/.mycroft/mycroft.conf"
        super().__init__(path)


class MycroftDefaultConfig(ReadOnlyConfig):
    def __init__(self):
        path = join(enclosure2rootdir(), "mycroft",
                    "configuration", "mycroft.conf")
        super().__init__(path)
        if not self.path or not isfile(self.path):
            LOG.warning("mycroft root path not found")

    def set_root_config_path(self, root_config):
        # in case we got it wrong / non standard
        self.path = root_config
        self.reload()


class MycroftSystemConfig(ReadOnlyConfig):
    def __init__(self, allow_overwrite=False):
        super().__init__(MYCROFT_SYSTEM_CONFIG, allow_overwrite)






from ovos_utils.log import LOG
from ovos_utils.messagebus import get_mycroft_bus, Message
import time
from os.path import isfile


class IntentApi:
    """
    Query Intent Service at runtime
    """

    def __init__(self, bus=None, timeout=5):
        self.bus = bus or get_mycroft_bus()
        self.timeout = timeout
        self.bus.on('intent.service.padatious.reply', self._receive_data)
        self.bus.on('intent.service.adapt.reply', self._receive_data)
        self.bus.on('intent.service.intent.reply', self._receive_data)
        self.bus.on('intent.service.skills.reply', self._receive_data)
        self.bus.on('intent.service.padatious.manifest', self._receive_data)
        self.bus.on('intent.service.adapt.manifest', self._receive_data)
        self.bus.on('intent.service.adapt.vocab.manifest', self._receive_data)
        self.bus.on('intent.service.padatious.entities.manifest',
                    self._receive_data)
        self._response = None
        self.waiting = False

    def _receive_data(self, message):
        self.waiting = False
        self._response = message.data

    def get_adapt_intent(self, utterance):
        """ get best adapt intent for utterance """
        start = time.time()
        self._response = None
        self.waiting = True
        self.bus.emit(Message("intent.service.adapt.get",
                              {"utterance": utterance},
                              context={"destination": "intent_service",
                                       "source": "intent_api"}))
        while self.waiting and time.time() - start <= self.timeout:
            time.sleep(0.3)
        if time.time() - start > self.timeout:
            LOG.error("Intent Service timed out!")
            return None
        return self._response["intent"]

    def get_padatious_intent(self, utterance):
        """ get best padatious intent for utterance """
        start = time.time()
        self._response = None
        self.waiting = True
        self.bus.emit(Message("intent.service.padatious.get",
                              {"utterance": utterance},
                              context={"destination": "intent_service",
                                       "source": "intent_api"}))
        while self.waiting and time.time() - start <= self.timeout:
            time.sleep(0.3)
        if time.time() - start > self.timeout:
            LOG.error("Intent Service timed out!")
            return None
        return self._response["intent"]

    def get_intent(self, utterance):
        """ get best intent for utterance """
        start = time.time()
        self._response = None
        self.waiting = True
        self.bus.emit(Message("intent.service.intent.get",
                              {"utterance": utterance},
                              context={"destination": "intent_service",
                                       "source": "intent_api"}))
        while self.waiting and time.time() - start <= self.timeout:
            time.sleep(0.3)
        if time.time() - start > self.timeout:
            LOG.error("Intent Service timed out!")
            return None
        return self._response["intent"]

    def get_skills(self):
        start = time.time()
        self._response = None
        self.waiting = True
        self.bus.emit(Message("intent.service.skills.get",
                              context={"destination": "intent_service",
                                       "source": "intent_api"}))
        while self.waiting and time.time() - start <= self.timeout:
            time.sleep(0.3)
        if time.time() - start > self.timeout:
            LOG.error("Intent Service timed out!")
            return None
        return self._response["skills"]

    def get_active_skills(self):
        start = time.time()
        self._response = None
        self.waiting = True
        self.bus.emit(Message("intent.service.active_skills.get",
                              context={"destination": "intent_service",
                                       "source": "intent_api"}))
        while self.waiting and time.time() - start <= self.timeout:
            time.sleep(0.3)
        if time.time() - start > self.timeout:
            LOG.error("Intent Service timed out!")
            return None
        return self._response["skills"]

    def get_adapt_manifest(self):
        start = time.time()
        self._response = None
        self.waiting = True
        self.bus.emit(Message("intent.service.adapt.manifest.get",
                              context={"destination": "intent_service",
                                       "source": "intent_api"}))
        while self.waiting and time.time() - start <= self.timeout:
            time.sleep(0.3)
        if time.time() - start > self.timeout:
            LOG.error("Intent Service timed out!")
            return None
        return self._response["intents"]

    def get_padatious_manifest(self):
        start = time.time()
        self._response = None
        self.waiting = True
        self.bus.emit(Message("intent.service.padatious.manifest.get",
                              context={"destination": "intent_service",
                                       "source": "intent_api"}))
        while self.waiting and time.time() - start <= self.timeout:
            time.sleep(0.3)
        if time.time() - start > self.timeout:
            LOG.error("Intent Service timed out!")
            return None
        return self._response["intents"]

    def get_intent_manifest(self):
        padatious = self.get_padatious_manifest()
        adapt = self.get_adapt_manifest()
        return {"adapt": adapt,
                "padatious": padatious}

    def get_vocab_manifest(self):
        start = time.time()
        self._response = None
        self.waiting = True
        self.bus.emit(Message("intent.service.adapt.vocab.manifest.get",
                              context={"destination": "intent_service",
                                       "source": "intent_api"}))
        while self.waiting and time.time() - start <= self.timeout:
            time.sleep(0.3)
        if time.time() - start > self.timeout:
            LOG.error("Intent Service timed out!")
            return None
        vocab = {}
        for voc in self._response["vocab"]:
            if voc.get("regex"):
                continue
            if voc["end"] not in vocab:
                vocab[voc["end"]] = {"samples": []}
            vocab[voc["end"]]["samples"].append(voc["start"])
        return [{"name": voc, "samples": vocab[voc]["samples"]}
                for voc in vocab]

    def get_regex_manifest(self):
        start = time.time()
        self._response = None
        self.waiting = True
        self.bus.emit(Message("intent.service.adapt.vocab.manifest.get",
                              context={"destination": "intent_service",
                                       "source": "intent_api"}))
        while self.waiting and time.time() - start <= self.timeout:
            time.sleep(0.3)
        if time.time() - start > self.timeout:
            LOG.error("Intent Service timed out!")
            return None

        vocab = {}
        for voc in self._response["vocab"]:
            if not voc.get("regex"):
                continue
            name = voc["regex"].split("(?P<")[-1].split(">")[0]
            if name not in vocab:
                vocab[name] = {"samples": []}
            vocab[name]["samples"].append(voc["regex"])
        return [{"name": voc, "regexes": vocab[voc]["samples"]}
                for voc in vocab]

    def get_entities_manifest(self):
        start = time.time()
        self._response = None
        self.waiting = True
        self.bus.emit(Message("intent.service.padatious.entities.manifest.get",
                              context={"destination": "intent_service",
                                       "source": "intent_api"}))
        while self.waiting and time.time() - start <= self.timeout:
            time.sleep(0.3)
        if time.time() - start > self.timeout:
            LOG.error("Intent Service timed out!")
            return None
        entities = []
        # read files
        for ent in self._response["entities"]:
            if isfile(ent["file_name"]):
                with open(ent["file_name"]) as f:
                    lines = f.read().replace("(", "").replace(")", "").split(
                        "\n")
                samples = []
                for l in lines:
                    samples += [a.strip() for a in l.split("|") if a.strip()]
                entities.append({"name": ent["name"], "samples": samples})
        return entities

    def get_keywords_manifest(self):
        padatious = self.get_entities_manifest()
        adapt = self.get_vocab_manifest()
        regex = self.get_regex_manifest()
        return {"adapt": adapt,
                "padatious": padatious,
                "regex": regex}




from ovos_utils.log import LOG

try:
    from google_trans_new import google_translator
except ImportError:
    google_translator = None

try:
    import pycld2 as cld2
except ImportError:
    cld2 = None

try:
    import cld3
except ImportError:
    cld3 = None


def code_to_name(lang_code):
    for name, code in cld2.LANGUAGES:
        if code == lang_code:
            return name.lower().capitalize()
    return "Unknown"


def detect_lang_naive(text, return_multiple=False, return_dict=False,
                      hint_language=None, filter_unreliable=False):
    """

    :param text:
    :param return_multiple bool if True return a list of all languages detected, else the top language
    :param return_dict: bool  if True returns all data, E.g.,  pt -> {'lang': 'Portuguese', 'lang_code': 'pt', 'conf': 0.96}
    :param hint_language: str  E.g., 'ITALIAN' or 'it' boosts Italian
    :return:
    """
    if cld2 is None:
        LOG.debug("run pip install pycld2")
        raise ImportError("pycld2 not installed")
    isReliable, textBytesFound, details = cld2.detect(text, hintLanguage=hint_language)
    languages = []

    # filter unreliable predictions
    if not isReliable and filter_unreliable:
        return None

    # select first language only
    if not return_multiple:
        details = [details[0]]

    for name, code, score, _ in details:
        if code == "un":
            continue
        if return_dict:
            languages.append({"lang": name.lower().capitalize(), "lang_code": code, "conf": score / 100})
        else:
            languages.append(code)

    # return top language only
    if not return_multiple:
        if not len(languages):
            return None
        return languages[0]
    return languages


def detect_lang_neural(text, return_multiple=False, return_dict=False,
                       hint_language=None, filter_unreliable=False):
    if cld3 is None:
        LOG.debug("run pip install pycld3")
        raise ImportError("pycld3 not installed")
    languages = []
    if return_multiple or hint_language:
        preds = sorted(cld3.get_frequent_languages(text, num_langs=5), key=lambda i: i.probability, reverse=True)
        for pred in preds:
            if filter_unreliable and not pred.is_reliable:
                continue
            if return_dict:
                languages += [{"lang_code": pred.language,
                               "lang": code_to_name(pred.language),
                               "conf": pred.probability}]
            else:
                languages.append(pred.language)

            if hint_language and hint_language == pred.language:
                languages = [languages[-1]]
                break
    else:
        pred = cld3.get_language(text)
        if filter_unreliable and not pred.is_reliable:
            pass
        elif return_dict:
            languages = [{"lang_code": pred.language,
                          "lang": code_to_name(pred.language),
                          "conf": pred.probability}]
        else:
            languages = [pred.language]

    # return top language only
    if not return_multiple:
        if not len(languages):
            return None
        return languages[0]
    return languages


def detect_lang_google(text, return_dict=False):
    if google_translator is None:
        LOG.debug("run pip install google_trans_new")
        raise ImportError("google_trans_new not installed")
    translator = google_translator()
    tx = translator.detect(text)
    if return_dict:
        return {"lang_code": tx[0], "lang": tx[1]}
    return tx[0]


def detect_lang(text, return_dict=False):
    if cld2 is not None:
        return detect_lang_naive(text, return_dict=return_dict)
    return detect_lang_google(text, return_dict=return_dict)


if __name__ == "__main__":
    assert detect_lang_google("olá eu chamo-me joaquim") == "pt"
    assert detect_lang_naive("olá eu chamo-me joaquim", return_dict=True) == \
           {'lang': 'Portuguese', 'lang_code': 'pt', 'conf': 0.96}

    assert detect_lang_naive("hello world") == "en"

    fr_en = """\
    France is the largest country in Western Europe and the third-largest in Europe as a whole.
    A accès aux chiens et aux frontaux qui lui ont été il peut consulter et modifier ses collections
    et exporter Cet article concerne le pays européen aujourd’hui appelé République française.
    Pour d’autres usages du nom France, Pour une aide rapide et effective, veuiller trouver votre aide
    dans le menu ci-dessus.
    Motoring events began soon after the construction of the first successful gasoline-fueled automobiles.
    The quick brown fox jumped over the lazy dog."""

    assert detect_lang_naive(fr_en) == "fr"
    assert detect_lang_naive(fr_en, return_multiple=True) == ['fr', 'en']

    assert detect_lang_naive(fr_en, hint_language="en", return_dict=True) == \
           {'lang': 'English', 'lang_code': 'en', 'conf': 0.6}  # boost english score

    assert detect_lang_naive(fr_en, return_dict=True, return_multiple=True)[1][
               "conf"] == 0.41  # compare conf with non boosted result

    assert detect_lang_naive(fr_en, return_dict=True) == \
           {'lang': 'French', 'lang_code': 'fr', 'conf': 0.58}
    assert detect_lang_neural(fr_en, return_dict=True) == \
           {'lang_code': 'fr', 'lang': 'French', 'conf': 0.9675191640853882}

    assert detect_lang_neural("This piece of text is in English. Този текст е на Български.",
                              return_dict=True,
                              return_multiple=True) == [
        {'lang_code': 'en', 'lang': 'English', 'conf': 0.9999790191650391},
        {'lang_code': 'bg', 'lang': 'Bulgarian', 'conf': 0.9173873066902161}
        ]

    assert detect_lang_neural("This piece of text is in English. Този текст е на Български.") == "en"
    assert detect_lang_neural("This piece of text is in English. Този текст е на Български.", return_dict=True, hint_language="bg") == \
           {'lang_code': 'bg', 'lang': 'Bulgarian', 'conf': 0.9173873066902161} # Boost bulgarian


# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from enum import IntEnum
from os.path import join, isfile
from ovos_utils.log import LOG
from ovos_utils import get_mycroft_root, resolve_ovos_resource_file
from ovos_utils.waiting_for_mycroft.settings_gui_generator import SettingsGuiGenerator

try:
    from mycroft.enclosure.gui import SkillGUI as _SkillGUI
    from mycroft.util import resolve_resource_file
    from mycroft.messagebus.message import Message
except ImportError:
    import sys
    MYCROFT_ROOT_PATH = get_mycroft_root()
    if MYCROFT_ROOT_PATH is not None:
        sys.path.append(MYCROFT_ROOT_PATH)
        from mycroft.enclosure.gui import SkillGUI as _SkillGUI
        from mycroft.util import resolve_resource_file
        from mycroft.messagebus.message import Message
    else:
        LOG.error("Could not find mycroft root path")
        raise ImportError

# implements the following GUI functionality
# https://github.com/MycroftAI/mycroft-core/pull/2683
# https://github.com/MycroftAI/mycroft-core/pull/2698


class GUIPlaybackStatus(IntEnum):
    STOPPED = 0
    PLAYING = 1
    PAUSED = 2
    UNDEFINED = 3


class SkillGUI(_SkillGUI):
    """SkillGUI - Interface to the Graphical User Interface

    Values set in this class are synced to the GUI, accessible within QML
    via the built-in sessionData mechanism.  For example, in Python you can
    write in a skill:
        self.gui['temp'] = 33
        self.gui.show_page('Weather.qml')
    Then in the Weather.qml you'd access the temp via code such as:
        text: sessionData.time
    """

    def __init__(self, skill):
        super().__init__(skill)
        self.video_info = None
        self.__skills_config = {}  # data object passed to skill's page
        self.settings_gui_generator = SettingsGuiGenerator()

    # video playback
    # https://github.com/MycroftAI/mycroft-core/pull/2683
    def setup_default_handlers(self):
        """Sets the handlers for the default messages."""
        super().setup_default_handlers()
        # Audio Service bus messages used by common play
        # TODO can we rename this namespace to mycroft.playback.XXX ?
        self.skill.add_event('mycroft.audio.service.pause',
                             self.__handle_gui_pause)
        self.skill.add_event('mycroft.audio.service.resume',
                             self.__handle_gui_resume)
        self.skill.add_event('mycroft.audio.service.stop',
                             self.__handle_gui_stop)
        self.skill.add_event('mycroft.audio.service.track_info',
                             self.__handle_gui_track_info)
        self.skill.add_event('mycroft.audio.queue_end',
                             self.__handle_gui_stop)
        self.skill.gui.register_handler('video.media.playback.ended',
                                        self.__handle_gui_stop)

    def __handle_gui_resume(self, message):
        """Resume video playback in gui"""
        self.resume_video()

    def __handle_gui_stop(self, message):
        """Stop video playback in gui"""
        self.stop_video()

    def __handle_gui_pause(self, message):
        """Pause video playback in gui"""
        self.pause_video()

    def __handle_gui_track_info(self, message):
        """Answer request information of current playing track.
         Needed for handling stop """
        if self.video_info:
            self.skill.bus.emit(
                message.reply('mycroft.audio.service.track_info_reply',
                              self.video_info))
        return self.video_info

    def play_video(self, url, title="", repeat=None, override_idle=True,
                   override_animations=None):
        """ Play video stream

        Arguments:
            url (str): URL of video source
            title (str): Title of media to be displayed
            repeat (boolean, int):
                True: Infinitly loops the current video track
                (int): Loops the video track for specified number of
                times.
            override_idle (boolean, int):
                True: Takes over the resting page indefinitely
                (int): Delays resting page for the specified number of
                       seconds.
            override_animations (boolean):
                True: Disables showing all platform skill animations.
                False: 'Default' always show animations.
        """
        self["playStatus"] = "play"
        self["video"] = url
        self["title"] = title
        self["playerRepeat"] = repeat
        self.video_info = {"title": title, "url": url}
        self.show_page("SYSTEM_VideoPlayer.qml",
                       override_idle=override_idle,
                       override_animations=override_animations)

    @property
    def is_video_displayed(self):
        """Returns whether the gui is in a video playback state.
        Eg if the video is paused, it would still be displayed on screen
        but the video itself is not "playing" so to speak"""
        return self.video_info is not None

    @property
    def playback_status(self):
        """Returns gui playback status,
        indicates if gui is playing, paused or stopped"""
        if self.__session_data.get("playStatus", -1) == "play":
            return GUIPlaybackStatus.PLAYING
        if self.__session_data.get("playStatus", -1) == "pause":
            return GUIPlaybackStatus.PAUSED
        if self.__session_data.get("playStatus", -1) == "stop":
            return GUIPlaybackStatus.STOPPED
        return GUIPlaybackStatus.UNDEFINED

    def pause_video(self):
        """Pause video playback."""
        if self.is_video_displayed:
            self["playStatus"] = "pause"

    def stop_video(self):
        """Stop video playback."""
        # TODO detect end of media playback from gui and call this
        if self.is_video_displayed:
            self["playStatus"] = "stop"
            self.clear()
        self.video_info = None

    def resume_video(self):
        """Resume paused video playback."""
        if self.__session_data.get("playStatus", "stop") == "pause":
            self["playStatus"] = "play"

    def shutdown(self):
        """Shutdown gui interface.

        Clear pages loaded through this interface and remove the skill
        reference to make ref counting warning more precise.
        """
        self.stop_video()
        super().shutdown()

    # skill settings
    # https://github.com/MycroftAI/mycroft-core/pull/2698
    def register_settings(self):
        """Register requested skill settings
        configuration in GUI.

        Registers handler to apply settings when
        updated via GUI interface.
        Register handler to update settings when
        updated via Web interface.
        """
        skill_id = self.skill.skill_id

        settingmeta_path = join(self.skill.root_dir,
                                "settingsmeta.json")
        if isfile(settingmeta_path):
            self.settings_gui_generator.populate(skill_id,
                                                 settingmeta_path,
                                                 self.skill.settings)
            apply_handler = skill_id + ".settings.set"
            update_handler = skill_id + ".settings.update"
            remove_pagehandler = skill_id + ".settings.remove_page"
            self.register_handler(apply_handler,
                                  self._apply_settings)
            self.register_handler(update_handler,
                                  self._update_settings)
            self.register_handler(remove_pagehandler,
                                  self._remove_settings_display)

        else:
            raise FileNotFoundError("Unable to find setting file for: {}".
                                    format(skill_id))

    def show_settings(self, override_idle=True,
                      override_animations=False):
        """Display skill configuration page in GUI.

        Arguments:
        override_idle (boolean, int):
                True: Takes over the resting page indefinitely
                (int): Delays resting page for the specified number of
                       seconds.
        override_animations (boolean):
                True: Disables showing all platform skill animations.
                False: 'Default' always show animations.
        """
        self.clear()
        self.__skills_config["sections"] = self.settings_gui_generator.fetch()
        self.__skills_config["skill_id"] = self.skill.skill_id
        self["skillsConfig"] = self.__skills_config
        self.show_page("SYSTEM_SkillSettings.qml",
                       override_idle=override_idle)

    def _apply_settings(self, message):
        """Store updated values for keys in skill settings.

        Arguments:
        message: Messagebus message
        """
        self.skill.settings[message.data["setting_key"]] = \
            message.data["setting_value"]

    def _update_settings(self, message):
        """Update changed skill settings in GUI.

        Arguments:
        message: Messagebus message
        """
        self.clear()
        self.settings_gui_generator.update(self.skill.settings)
        self.show_settings()

    def _remove_settings_display(self, message):
        """Removes skill settings page from GUI.

        Arguments:
        message: Messagebus message
        """
        self.clear()
        self.remove_page("SYSTEM_SkillSettings.qml")

    # everything above is a new method, everything bellow is partial overrides
    # these unfortunately require implementing the full method, may get out
    # of sync over time
    def show_pages(self, page_names, index=0, override_idle=None,
                   override_animations=False):
        """Begin showing the list of pages in the GUI.

        Arguments:
            page_names (list): List of page names (str) to display, such as
                               ["Weather.qml", "Forecast.qml", "Details.qml"]
            index (int): Page number (0-based) to show initially.  For the
                         above list a value of 1 would start on "Forecast.qml"
            override_idle (boolean, int):
                True: Takes over the resting page indefinitely
                (int): Delays resting page for the specified number of
                       seconds.
            override_animations (boolean):
                True: Disables showing all platform skill animations.
                False: 'Default' always show animations.
        """
        if not isinstance(page_names, list):
            raise ValueError('page_names must be a list')

        if index > len(page_names):
            raise ValueError('Default index is larger than page list length')

        self.page = page_names[index]

        # First sync any data...
        data = self.__session_data.copy()
        data.update({'__from': self.skill.skill_id})
        self.skill.bus.emit(Message("gui.value.set", data))

        # Convert pages to full reference
        page_urls = []
        for name in page_names:

            if name.startswith("SYSTEM"):
                page = resolve_resource_file(join('ui', name))
            else:
                page = self.skill.find_resource(name, 'ui')

            if not page:
                # override to look for bundled pages
                page = resolve_ovos_resource_file(join('ui', name)) or \
                        resolve_ovos_resource_file(name)
            if page:
                if self.config.get('remote'):
                    page_urls.append(self.remote_url + "/" + page)
                else:
                    page_urls.append("file://" + page)
            else:
                raise FileNotFoundError("Unable to find page: {}".format(name))

        self.skill.bus.emit(Message("gui.page.show",
                                    {"page": page_urls,
                                     "index": index,
                                     "__from": self.skill.skill_id,
                                     "__idle": override_idle,
                                     "__animations": override_animations}))

    def remove_pages(self, page_names):
        """Remove a list of pages in the GUI.

        Arguments:
            page_names (list): List of page names (str) to display, such as
                               ["Weather.qml", "Forecast.qml", "Other.qml"]
        """
        if not isinstance(page_names, list):
            raise ValueError('page_names must be a list')

        # Convert pages to full reference
        page_urls = []
        for name in page_names:
            if name.startswith("SYSTEM"):
                page = resolve_resource_file(join('ui', name))
            else:
                page = self.skill.find_resource(name, 'ui')

            if not page:
                # override to look for bundled pages
                page = resolve_ovos_resource_file(join('ui', name)) or \
                       resolve_ovos_resource_file(name)

            if page:
                page_urls.append("file://" + page)
            else:
                raise FileNotFoundError("Unable to find page: {}".format(name))

        self.skill.bus.emit(Message("gui.page.delete",
                                    {"page": page_urls,
                                     "__from": self.skill.skill_id}))




from ovos_utils.configuration import read_mycroft_config, update_mycroft_config
from os.path import join


def blacklist_skill(skill):
    skills_config = read_mycroft_config().get("skills", {})
    blacklisted_skills = skills_config.get("blacklisted_skills", [])
    if skill not in blacklisted_skills:
        blacklisted_skills.append(skill)
        conf = {
            "skills": {
                "blacklisted_skills": blacklisted_skills
            }
        }
        update_mycroft_config(conf)
        return True
    return False


def whitelist_skill(skill):
    skills_config = read_mycroft_config().get("skills", {})
    blacklisted_skills = skills_config.get("blacklisted_skills", [])
    if skill in blacklisted_skills:
        blacklisted_skills.pop(skill)
        conf = {
            "skills": {
                "blacklisted_skills": blacklisted_skills
            }
        }
        update_mycroft_config(conf)
        return True
    return False


def make_priority_skill(skill):
    skills_config = read_mycroft_config().get("skills", {})
    priority_skills = skills_config.get("priority_skills", [])
    if skill not in priority_skills:
        priority_skills.append(skill)
        conf = {
            "skills": {
                "priority_skills": priority_skills
            }
        }
        update_mycroft_config(conf)
        return True
    return False


def get_skills_folder():
    config = read_mycroft_config()
    data_dir = config["data_dir"]
    skill_folder = config["skills"]["msm"]["directory"]
    return join(data_dir, skill_folder)



#!/usr/bin/env python3
import sys
import os
import subprocess
import re
from enum import Enum
import platform
from ovos_utils.display import can_display, is_gui_installed


class MycroftRootLocations(str, Enum):
    PICROFT = "/home/pi/mycroft-core"
    BIGSCREEN = "/home/mycroft/mycroft-core"
    OVOS = "/home/mycroft/mycroft-core"  # TODO ovos here
    OLD_MARK1 = "/opt/venvs/mycroft-core/lib/python3.4/site-packages/"
    MARK1 = "/opt/venvs/mycroft-core/lib/python3.7/site-packages/"
    MARK2 = "/home/mycroft/mycroft-core"  # TODO mark2 here


def search_mycroft_core_location():
    for p in MycroftRootLocations:
        if os.path.isdir(p):
            return p
    return None


def get_desktop_environment():
    # From http://stackoverflow.com/questions/2035657/what-is-my-current-desktop-environment
    # and http://ubuntuforums.org/showthread.php?t=652320
    # and http://ubuntuforums.org/showthread.php?t=652320
    # and http://ubuntuforums.org/showthread.php?t=1139057
    if sys.platform in ["win32", "cygwin"]:
        return "windows"
    elif sys.platform == "darwin":
        return "mac"
    else:  # Most likely either a POSIX system or something not much common
        desktop_session = os.environ.get("DESKTOP_SESSION")
        if desktop_session is not None:  # easier to match if we doesn't have  to deal with character cases
            desktop_session = desktop_session.lower()
            if desktop_session in ["gnome", "unity", "cinnamon", "mate", "xfce4", "lxde", "fluxbox",
                                   "blackbox", "openbox", "icewm", "jwm", "afterstep", "trinity", "kde"]:
                return desktop_session
            # Special cases
            # Canonical sets $DESKTOP_SESSION to Lubuntu rather than LXDE if using LXDE.
            # There is no guarantee that they will not do the same with the other desktop environments.
            elif "xfce" in desktop_session or desktop_session.startswith("xubuntu"):
                return "xfce4"
            elif desktop_session.startswith("ubuntu"):
                return "unity"
            elif desktop_session.startswith("lubuntu"):
                return "lxde"
            elif desktop_session.startswith("kubuntu"):
                return "kde"
            elif desktop_session.startswith("razor"):  # e.g. razorkwin
                return "razor-qt"
            elif desktop_session.startswith("wmaker"):  # e.g. wmaker-common
                return "windowmaker"
        if os.environ.get('KDE_FULL_SESSION') == 'true':
            return "kde"
        elif os.environ.get('GNOME_DESKTOP_SESSION_ID'):
            if not "deprecated" in os.environ.get('GNOME_DESKTOP_SESSION_ID'):
                return "gnome2"
        # From http://ubuntuforums.org/showthread.php?t=652320
        elif is_process_running("xfce-mcs-manage"):
            return "xfce4"
        elif is_process_running("ksmserver"):
            return "kde"
        elif is_process_running("icewm"):
            return "icewm"
        elif is_process_running("fluxbox"):
            return "fluxbox"
        elif is_process_running("jwm"):
            return "jwm"
    return "unknown"


def is_process_running(process):
    # From http://www.bloggerpolis.com/2011/05/how-to-check-if-a-process-is-running-using-python/
    # and http://richarddingwall.name/2009/06/18/windows-equivalents-of-ps-and-kill-commands/
    try:  # Linux/Unix
        s = subprocess.Popen(["ps", "axw"], stdout=subprocess.PIPE)
    except:  # Windows
        s = subprocess.Popen(["tasklist", "/v"], stdout=subprocess.PIPE)
    for x in s.stdout:
        if re.search(process, x.decode("utf-8")):
            return True
    return False


def get_platform_fingerprint():
    # TODO hostname
    return {
        "platform": platform.platform(),
        "python_version": platform.python_version(),
        "system": platform.system(),
        "version": platform.version(),
        "arch": platform.machine(),
        "release": platform.release(),
        "node": platform.node(),
        "desktop_env": get_desktop_environment(),
        "mycroft_core_location": search_mycroft_core_location(),
        "can_display": can_display(),
        "is_gui_installed": is_gui_installed(),
        "pulseaudio_running": is_process_running("pulseaudio")
    }


def ntp_sync():
    # Force the system clock to synchronize with internet time servers
    subprocess.call('service ntp stop', shell=True)
    subprocess.call('ntpd -gq', shell=True)
    subprocess.call('service ntp start', shell=True)


def system_shutdown():
    # Turn the system completely off (with no option to inhibit it)
    subprocess.call('sudo systemctl poweroff -i', shell=True)


def system_reboot():
    # Shut down and restart the system
    subprocess.call('sudo systemctl reboot -i', shell=True)


def ssh_enable():
    # Permanently allow SSH access
    subprocess.call('sudo systemctl enable ssh.service', shell=True)
    subprocess.call('sudo systemctl start ssh.service', shell=True)


def ssh_disable():
    # Permanently block SSH access from the outside
    subprocess.call('sudo systemctl stop ssh.service', shell=True)
    subprocess.call('sudo systemctl disable ssh.service', shell=True)



import os
import subprocess


def can_display():
    have_display = "DISPLAY" in os.environ
    if not have_display:
        # fallback check using matplotlib if available
        try:
            import matplotlib.pyplot as plt
            try:
                plt.figure()
                have_display = True
            except:
                have_display = False
        except ImportError:
            pass
    return have_display


def is_gui_installed():
    return bool(subprocess.check_output(
        "which mycroft-gui-app", shell=True).decode("utf-8").strip())


def is_gui_connected():
    # TODO bus api for https://github.com/MycroftAI/mycroft-core/pull/2682
    return False



import unittest
from copy import deepcopy
from ovos_utils.json_helper import merge_dict


class TestJsonHelpers(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.base_dict = {"one": 1, "two": 2, "three": 3,
                          "four": ["foo", "bar", "baz"], "five": 50}
        self.delta_dict = {"two": 2, "three": 30,
                           "four": [4, 5, 6, "foo"], "five": None}

    def test_merge_dict(self):
        self.assertEqual(
            merge_dict(deepcopy(self.base_dict), deepcopy(self.delta_dict)),
            {"one": 1, "two": 2, "three": 30, "four": [
                4, 5, 6, "foo"], "five": None}
        )

    def test_merge_dict_lists(self):

        self.assertEqual(
            merge_dict(deepcopy(self.base_dict), deepcopy(self.delta_dict),
                       merge_lists=True, no_dupes=False),
            {"one": 1, "two": 2, "three": 30, "four": [
                "foo", "bar", "baz", 4, 5, 6, "foo"], "five": None}
        )

    def test_merge_dict_skip_empty(self):

        self.assertEqual(
            merge_dict(deepcopy(self.base_dict), deepcopy(self.delta_dict),
                       merge_lists=True, skip_empty=True, no_dupes=False),
            {"one": 1, "two": 2, "three": 30, "four": [
                "foo", "bar", "baz", 4, 5, 6, "foo"], "five": 50}
        )

    # Difficult to test because order is not currently guaranteed.

    def test_merge_dict_no_dupes(self):

        self.assertEqual(
            merge_dict(deepcopy(self.base_dict), deepcopy(self.delta_dict), merge_lists=True,
                       skip_empty=True, no_dupes=True),
            {"one": 1, "two": 2, "three": 30, "four": [
                "foo", "bar", "baz", 4, 5, 6], "five": 50}
        )



from setuptools import setup

setup(
    name='ovos_utils',
    version='0.0.4rc2',
    packages=['ovos_utils',
              'ovos_utils.waiting_for_mycroft',
              'ovos_utils.misc',
              'ovos_utils.intents',
              'ovos_utils.sound',
              "ovos_utils.enclosure",
              'ovos_utils.enclosure.mark1',
              'ovos_utils.enclosure.mark1.eyes',
              'ovos_utils.enclosure.mark1.faceplate',
              'ovos_utils.skills',
              'ovos_utils.skills.templates',
              'ovos_utils.lang'],
    url='https://github.com/OpenVoiceOS/ovos_utils',
    install_requires=["phoneme_guesser",
                      "mycroft-messagebus-client",
                      "inflection",
                      "colour",
                      "pexpect",
                      "PyYAML",
                      "json_database",
                      "requests"],
    include_package_data=True,
    license='Apache',
    author='jarbasAI',
    author_email='jarbasai@mailfence.com',
    description='collection of simple utilities for use across the mycroft ecosystem'
)



from ovos_utils.waiting_for_mycroft.common_play import CommonPlaySkill, \
    CPSMatchLevel, CPSTrackStatus, CPSMatchType
from os.path import join, dirname, basename
import random
import re
import datetime
from ovos_utils import get_mycroft_root
from ovos_utils.log import LOG
from ovos_utils.parse import fuzzy_match

try:
    from mycroft.skills.core import intent_file_handler
except ImportError:
    import sys
    MYCROFT_ROOT_PATH = get_mycroft_root()
    if MYCROFT_ROOT_PATH is not None:
        sys.path.append(MYCROFT_ROOT_PATH)
        from mycroft.skills.core import intent_file_handler
    else:
        LOG.error("Could not find mycroft root path")
        raise ImportError


try:
    from json_database import JsonStorageXDG
except ImportError:
    LOG.error("json_database not installed!")
    LOG.debug("pip install json_database>=0.2.1")


try:
    from pyvod import Collection, Media
except ImportError:
    LOG.error("py_VOD not installed!")
    LOG.debug("py_VOD>=0.3.1")


def datestr2ts(datestr):
    y = int(datestr[:4])
    m = int(datestr[4:6])
    d = int(datestr[-2:])
    dt = datetime.datetime(y, m, d)
    return dt.timestamp()


class MediaCollectionSkill(CommonPlaySkill):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "max_videos" not in self.settings:
            self.settings["max_videos"] = 500
        if "min_duration" not in self.settings:
            self.settings["min_duration"] = 130
        if "max_duration" not in self.settings:
            self.settings["max_duration"] = -1
        if "shuffle_menu" not in self.settings:
            self.settings["shuffle_menu"] = False
        if "filter_live" not in self.settings:
            self.settings["filter_live"] = False
        default_logo = "https://github.com/OpenVoiceOS/ovos_assets/raw" \
                       "/master/Logo/ovos-logo-mono-256.png"
        self.supported_media = [CPSMatchType.GENERIC, CPSMatchType.VIDEO]
        db_path = join(dirname(__file__), "res", self.name + ".jsondb")
        self.message_namespace = basename(dirname(__file__)) + ".ovos_utils"
        self.media_collection = Collection(self.name,
                                           logo=default_logo,
                                           db_path=db_path)

    def initialize(self):
        self.initialize_media_commons()

    def initialize_media_commons(self):
        # generic ovos events
        self.gui.register_handler("ovos_utils.play_event",
                                  self.play_video_event)
        self.gui.register_handler("ovos_utils.clear_history",
                                  self.handle_clear_history)

        # skill specific events
        self.add_event('{msg_base}.home'.format(msg_base=self.message_namespace),
                       self.handle_homescreen)
        self.gui.register_handler(
            "{msg_base}.play_event".format(msg_base=self.message_namespace),
            self.play_video_event)
        self.gui.register_handler(
            "{msg_base}.clear_history".format(msg_base=self.message_namespace),
            self.handle_clear_history)

    @property
    def videos(self):
        try:
            # load video catalog
            videos = [ch.as_json() for ch in self.media_collection.entries]
            # set skill_id
            for idx, v in enumerate(videos):
                videos[idx]["skill"] = self.skill_id
            # return sorted
            return self.sort_videos(videos)
        except Exception as e:
            return []

    # homescreen / menu
    def sort_videos(self, videos):
        # this will filter private videos
        videos = [v for v in videos if v.get("upload_date")]

        # this will filter live videos
        live = [v for v in videos if v.get("is_live")]
        videos = [v for v in videos if not v.get("is_live")]

        # sort by upload date
        videos = sorted(videos,
                        key=lambda kv: datestr2ts(kv["upload_date"]),
                        reverse=True)

        # live streams before videos
        if not self.settings["filter_live"]:
            videos = live + videos

        return videos

    def filter_videos(self):
        # filter by duration
        videos = [v for v in self.videos if int(v.get("duration", 0)) >=
                  self.settings["min_duration"]]
        if self.settings["max_duration"] > 0:
            videos = [v for v in videos if int(v.get("duration", 0)) <=
                      self.settings["max_duration"]]

        # TODO filter behind the scenes, clips etc based on
        #  title/tags/description/keywords required or forbidden

        if self.settings["shuffle_menu"]:
            random.shuffle(videos)

        if self.settings["max_videos"]:
            # rendering takes forever if there are too many entries
            videos = videos[:self.settings["max_videos"]]

        return videos

    def handle_homescreen(self, message):
        self.gui.clear()
        self.gui["videosHomeModel"] = self.filter_videos()
        self.gui["historyModel"] = JsonStorageXDG("{msg_base}.history".format(msg_base=self.message_namespace)) \
            .get("model", [])
        self.gui.show_page("Homescreen.qml", override_idle=True)

    def play_video_event(self, message):
        video_data = message.data["modelData"]
        if video_data["skill_id"] == self.skill_id:
            self.play_video(video_data)

    # watch history database
    def add_to_history(self, video_data):
        # History
        historyDB = JsonStorageXDG("{msg_base}.history".format(msg_base=self.message_namespace))
        if "model" not in historyDB:
            historyDB["model"] = []
        historyDB["model"].append(video_data)
        historyDB.store()
        self.gui["historyModel"] = historyDB["model"]

    def handle_clear_history(self, message):
        video_data = message.data["modelData"]
        if video_data["skill_id"] == self.skill_id:
            historyDB = JsonStorageXDG("{msg_base}.history"
                                       .format(msg_base=self.message_namespace))
            historyDB["model"] = []
            historyDB.store()

    # matching
    def match_media_type(self, phrase, media_type):
        match = None
        score = 0

        if media_type == CPSMatchType.VIDEO:
            score += 0.05
            match = CPSMatchLevel.GENERIC

        return match, score

    def match_tags(self, video, phrase, match):
        score = 0
        # score tags
        leftover_text = phrase
        tags = list(set(video.get("tags") or []))
        if tags:
            # tag match bonus
            for tag in tags:
                tag = tag.lower().strip()
                if tag in phrase:
                    match = CPSMatchLevel.CATEGORY
                    score += 0.05
                    leftover_text = leftover_text.replace(tag, "")
        return match, score, leftover_text

    def match_description(self, video, phrase, match):
        # score description
        score = 0
        leftover_text = phrase
        words = video.get("description", "").split(" ")
        for word in words:
            if len(word) > 4 and word in self.normalize_title(leftover_text):
                score += 0.05
                leftover_text = leftover_text.replace(word, "")
        return match, score, leftover_text

    def match_title(self, videos, phrase, match):
        # match video name
        clean_phrase = self.normalize_title(phrase)
        leftover_text = phrase
        best_score = 0
        best_video = random.choice(videos)
        for video in videos:
            title = video["title"]
            score = fuzzy_match(clean_phrase, self.normalize_title(title))
            if phrase.lower() in title.lower() or \
                    clean_phrase in self.normalize_title(title):
                score += 0.3
            if score >= best_score:
                # TODO handle ties
                match = CPSMatchLevel.TITLE
                best_video = video
                best_score = score
                leftover_text = phrase.replace(title, "")
        return match, best_score, best_video, leftover_text

    def remove_voc(self, utt, voc_filename, lang=None):
        lang = lang or self.lang
        cache_key = lang + voc_filename

        if cache_key not in self.voc_match_cache:
            # trigger caching
            self.voc_match(utt, voc_filename, lang)

        if utt:
            # Check for matches against complete words
            for i in self.voc_match_cache[cache_key]:
                # Substitute only whole words matching the token
                utt = re.sub(r'\b' + i + r"\b", "", utt)

        return utt

    def normalize_title(self, title):
        title = title.lower().strip()
        title = self.remove_voc(title, "video")
        title = title.replace("|", "").replace('"', "") \
            .replace(':', "").replace('”', "").replace('“', "") \
            .strip()
        return " ".join([w for w in title.split(" ") if w])  # remove extra
        # spaces

    # common play
    def calc_final_score(self, phrase, base_score, match_level):
        return base_score, match_level

    def base_CPS_match(self, phrase, media_type):
        best_score = 0
        # see if media type is in query, base_score will depend if "video" is in query
        match, base_score = self.match_media_type(phrase, media_type)
        videos = list(self.videos)
        best_video = random.choice(self.videos)
        # match video data
        scores = []
        for video in videos:
            match, score, _ = self.match_tags(video, phrase, match)
            # match, score, leftover_text = self.match_description(video, leftover_text, match)
            scores.append((video, score))
            if score > best_score:
                best_video = video
                best_score = score

        self.log.debug("Best Tags Match: {s}, {t}".format(
            s=best_score, t=best_video["title"]))

        # match video name
        match, title_score, best_title, leftover_text = self.match_title(
            videos, phrase, match)
        self.log.debug("Best Title Match: {s}, {t}".format(
            s=title_score, t=best_title["title"]))

        # title more important than tags
        if title_score + 0.15 > best_score:
            best_video = best_title
            best_score = title_score

        # sort matches
        scores = sorted(scores, key=lambda k: k[1], reverse=True)
        scores.insert(0, (best_title, title_score))
        scores.remove((best_video, best_score))
        scores.insert(0, (best_video, best_score))

        # choose from top N
        if best_score < 0.5:
            n = 50
        elif best_score < 0.6:
            n = 10
        elif best_score < 0.8:
            n = 3
        else:
            n = 1

        candidates = scores[:n]
        self.log.info("Choosing randomly from top {n} Dust matches".format(
            n=len(candidates)))
        best_video = random.choice(candidates)[0]

        # calc final confidence
        score = base_score + best_score
        score = self.calc_final_score(phrase, score, match)
        if isinstance(score, float):
            if score >= 0.85:
                match = CPSMatchLevel.EXACT
            elif score >= 0.7:
                match = CPSMatchLevel.MULTI_KEY
            elif score >= 0.5:
                match = CPSMatchLevel.TITLE
        else:
            score, match = score

        self.log.debug("Best video: " + best_video["title"])

        if match is not None:
            return (leftover_text, match, best_video)
        return None

    def CPS_match_query_phrase(self, phrase, media_type):
        match = self.base_CPS_match(phrase, media_type)
        if match is None:
            return None
        # match == (leftover_text, CPSMatchLevel, best_video_data)
        return match

    def CPS_start(self, phrase, data):
        self.play_video(data)

    def play_video(self, video_data):
        # TODO audio only
        # if self.gui.connected:
        #    ...
        # else:
        #    self.audioservice.play(video_data["url"])
        # add to playback history

        self.add_to_history(video_data)
        # play video
        video = Media.from_json(video_data)
        url = str(video.streams[0])
        self.gui.play_video(url, video.name)


def create_skill():
    return MediaCollectionSkill()



from ovos_utils.system import MycroftRootLocations
from enum import Enum
from os.path import exists


class MycroftEnclosures(str, Enum):
    PICROFT = "picroft"
    BIGSCREEN = "mycroft_mark_2"  # TODO handle bigscreen
    OVOS = "OpenVoiceOS"
    OLD_MARK1 = "mycroft_mark_1(old)"
    MARK1 = "mycroft_mark_1"
    MARK2 = "mycroft_mark_2"
    OTHER = "unknown"


def enclosure2rootdir(enclosure=None):
    enclosure = enclosure or detect_enclosure()
    if enclosure == MycroftEnclosures.OLD_MARK1:
        return MycroftRootLocations.OLD_MARK1
    elif enclosure == MycroftEnclosures.MARK1:
        return MycroftRootLocations.MARK1
    elif enclosure == MycroftEnclosures.MARK2:
        return MycroftRootLocations.MARK2
    elif enclosure == MycroftEnclosures.PICROFT:
        return MycroftRootLocations.PICROFT
    elif enclosure == MycroftEnclosures.OVOS:
        return MycroftRootLocations.OVOS
    elif enclosure == MycroftEnclosures.BIGSCREEN:
        return MycroftRootLocations.BIGSCREEN
    raise EnvironmentError


def detect_enclosure():
    # TODO avoid circular import better
    # this is the only safe import to use from ovos_utils.configuration,
    # other methods are NOT, because they depend on this method directly,
    # ie, configuration objects use detect_enclosure
    from ovos_utils.configuration import MycroftSystemConfig

    # TODO very naive check, improve this
    # use db of reference fingerprints
    fingerprint = MycroftSystemConfig().get("enclosure", {}).get("platform",
                                                                 "unknown")
    if fingerprint == "OpenVoiceOS":
        return MycroftEnclosures.OVOS
    elif fingerprint == "mycroft_mark_1":
        if exists(MycroftRootLocations.OLD_MARK1):
            return MycroftEnclosures.OLD_MARK1
        return MycroftEnclosures.MARK1
    elif fingerprint == "mycroft_mark_2":
        # TODO bigscreen also reports this...
        return MycroftEnclosures.MARK2
    elif fingerprint == "picroft":
        return MycroftEnclosures.PICROFT

    return MycroftEnclosures.OTHER



# Copyright 2020 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import yaml
import pathlib


class SettingsGuiGenerator:
    """Skill Settings Generator For GUI. """

    def __init__(self):
        """ Create a SettingList Object """
        self.settings_list = []

    def populate(self, skill_id, settings_file, settings_dict):
        """
        Populates settings list for current skill.

        Arguments:
            skill_id: ID of target skill.
            settings_file: Settings meta file from skill folder.
            settings_dict: Dictionary of current settings.json file.
        """
        file_type = pathlib.Path(settings_file).suffix

        if file_type == ".json":
            with open(settings_file, 'r') as f:
                settingsmeta_dict = json.load(f)

                __skillMetaData = settingsmeta_dict.get('skillMetadata')
                for section in __skillMetaData.get('sections'):
                    self.settings_list.append(section)

        if file_type == ".yaml":
            with open(settings_file, 'r') as f:
                settingsmeta_dict = yaml.safe_load(f)

                __skillMetaData = settingsmeta_dict.get('skillMetadata')
                for section in __skillMetaData.get('sections'):
                    self.settings_list.append(section)

        if settings_dict is not None:
            __updated_list = []
            for sections in self.settings_list:
                for fields in sections['fields']:
                    if "name" in fields:
                        if fields["name"] in settings_dict.keys():
                            fields["value"] = settings_dict[fields["name"]]

                __updated_list.append(sections)

            self.clear()
            self.settings_list = __updated_list

    def fetch(self):
        """Return Settings List """
        return self.settings_list

    def clear(self):
        """Clear Settings List """
        self.settings_list.clear()

    def update(self, settings_dict):
        """Getting Changed Settings & Update List.

        Arguments:
        settings_dict: Dictionary of current settings.json file.
        """
        __updated_list = []
        for sections in self.settings_list:
            for fields in sections['fields']:
                if "name" in fields:
                    if fields["name"] in settings_dict.keys():
                        fields["value"] = settings_dict[fields["name"]]

            __updated_list.append(sections)

        self.clear()
        self.settings_list = __updated_list


from os.path import expanduser, isdir, dirname
from os import system, makedirs


def get_tts(sentence, lang="en-us", mp3_file="/tmp/google_tx_tts.mp3"):
    # TODO privacy issues - https://github.com/OpenVoiceOS/ovos_utils/issues/2
    ext = "mp3"
    if not mp3_file.endswith(ext):
        mp3_file += "." + ext
    mp3_file = expanduser(mp3_file)
    if not isdir(dirname(mp3_file)):
        makedirs(dirname(mp3_file))
    get_sentence = 'wget -q -U Mozilla -O' + mp3_file + \
                   ' "https://translate.google.com/translate_tts?tl=' + \
                   lang + '&q=' + sentence + '&client=tw-ob' + '"'
    system(get_sentence)
    return mp3_file
